// Define the needed variables and objects
var yetify = require('yetify'),
        config = require('getconfig'),
        uuid = require('node-uuid'),
        crypto = require('crypto'),
        fs = require('fs'),
        port = parseInt(process.env.PORT || config.server.port, 10),
        server = null,
        winston = require('winston'),
        roomTokens = {},
        tokenList = {},
        fileServer = null,
        serverHandler = null,
        adminAccess = null,
        urlParser = require('url'),
        protocol = 'http',
        options = {},
        tokensEnabled = true,
        ersVersion = require('./package.json').version;

var logger = new (winston.Logger)({
    transports: [
        new (winston.transports.Console)({level: 'debug', colorize: true}),
        new (winston.transports.File)({filename: 'ers.log',
            level: 'debug',
            json: false,
            tailable: true})
    ]
});

if (config.server.webServerEnabled) {
    // Instantiate a copy of static webserver
    fileServer = new ((require('node-static')).Server)();

    serverHandler = function (req, res) {
        fileServer.serve(req, res);
        logger.log('debug', 'Client request: [' + req.connection.remoteAddress + ']');
    };
} else {
    // No webserver
    serverHandler = function (req, res) {
        res.writeHead(404);
        res.end();
    };
}

// Create an http(s) server instance to which socket.io can listen to
// Adjust the settings if we are using https
if (config.server.secure) {
    protocol = 'https';
    options = {
        key: fs.readFileSync(config.server.key),
        cert: fs.readFileSync(config.server.cert),
        passphrase: config.server.password
    };
    
    server = require(protocol).createServer(options, serverHandler);
} else {
    server = require(protocol).createServer(serverHandler);
}

server.listen(port);

var io = require('socket.io').listen(server);

// Set the log level of IO sockets to higher than debug
io.set('log level', 2);

// Instantiate service to listen to admin control
if (config.server.adminAccessEnabled) {
    var adminHandler = function (req, res) {
        logger.log('info', 'Client request: [' + req.connection.remoteAddress + ']');
        if ((req.connection.remoteAddress).indexOf(config.server.adminIP) > -1) {
            // Access to admin settings is allowed

            // Check if this is a request for the admin page
            if ((req.url).indexOf('?') === -1) {
                fileServer.serve(req, res);
            } else {
                // Handle the message normally
                handleMessage(req, res);
            }
        } else {
            // Invalid IP address of admin access
            res.writeHead(404);
            res.end();
        }
    };

    // Adjust the settings if we are using https
    if (config.server.secure) {
        adminAccess = require(protocol).createServer(options, adminHandler);
    } else {
        adminAccess = require(protocol).createServer(adminHandler);
    }

    adminAccess.listen(config.server.adminPort);
}

// Set the flag indicating the usage of tokens
if (config.server.tokensEnabled === true) {
    // Fire up a timer that triggers every minute to clean-up token list
    setInterval(function () {
        var now = Date.now();
        for (var token in tokenList) {
            if (tokenList[token].used && ((now - tokenList[token].time) > 600000)) {
                // This token has been marked used and has exceeded 10mins
                delete tokenList[token];
                logger.log('debug', "Token (%s) has been deleted.", token);
            }
        }
    }, 60000);
} else {
    tokensEnabled = false;
}

/**
 * Handle message with the following request format:
 * http(s)://url:port/?c=commandName&i=Id
 * 
 * where:
 *   commanName can either be addToken, delToken, listTokens, addRoom, delRoom, listRooms
 *   Id is the ID of the token or room to be added or deleted
 * 
 * listTokens and listRooms command does not require any argument
 * 
 * @param {type} request
 * @param {type} response
 * @returns {undefined}
 */
function handleMessage(request, response) {
    //TODO: should we go with post? Get would suffice I guess since we only have
    // two parameters
    var query = urlParser.parse(request.url, true).query;
    var url = request.url;
    var ip = request.connection.remoteAddress;

    // Check if we have a valid command
    if ((typeof query.c === 'undefined') || (query.c === '')) {
        logInvalidRequest(url, ip);
        return;
    }

    // Check if we have the needed parameter
    if (((query.c === 'addToken') || (query.c === 'addRoom') ||
            (query.c === 'delToken') || (query.c === 'delRoom')) &&
            ((typeof query.i === 'undefined') || (query.i === ''))) {
        logInvalidRequest(url, ip);
        return;
    }

    // Process the command
    var status = 'SUCCESS';
    var description = '';
    var data = '';
    switch (query.c) {
        case 'addToken':
            if (addToken(query.i)) {
                description = 'Token added successfully.';
            } else {
                description = 'Token ' + query.i + ' already exists!';
                status = 'FAIL';
            }
            break;
        case 'delToken':
            if (delToken(query.i)) {
                description = 'Token deleted successfully.';
            } else {
                description = 'Token ' + query.i + ' not found!';
                status = 'FAIL';
            }
            break;
        case 'listTokens':
            data = listTokens();
            break;
        case 'addRoom':
            addRoom(query.i);
            description = 'Room added successfully.';
            break;
        case 'delRoom':
            if (delRoom(query.i)) {
                description = 'Room deleted successfully.';
            } else {
                description = 'Room ' + query.i + ' not found!';
                status = 'FAIL';
            }
            break;
        case 'listRooms':
            data = listRooms();
            break;
        case 'addEmsToken':
            if (addEmsToken(query.i)) {
                description = 'EMS token added successfully.';
            } else {
                description = 'EMS token ' + query.i + ' already exists!';
                status = 'FAIL';
            }
            break;
        case 'delEmsToken':
            if (delEmsToken(query.i)) {
                description = 'EMS token deleted successfully.';
            } else {
                description = 'EMS token ' + query.i + ' not found!';
                status = 'FAIL';
            }
            break;
        case 'listEmsTokens':
            data = listEmsTokens();
            break;
        case 'version':
            data = version();
            break;
        default:
            logInvalidRequest(url, ip);
            description = 'Invalid request!';
            status = 'FAIL';
            break;
    }

    // Process the response
    var res = JSON.stringify({
        url: url,
        status: status,
        description: description,
        data: data
    });

    response.writeHead(200, {"Content-Type": "application/json"});
    response.end(res);
}

function logInvalidRequest(url, ip) {
    logger.log('error', "Invalid request (%s) from: %s", url, ip);
}

function logInvalidToken(token, ip) {
    logger.log('error', "Invalid token (%s) from: %s", token, ip);
}

function addToken(tokenId) {
    if (tokenId in tokenList) {
        logger.log('warn', "Token (%s) already exists!", tokenId);
        return false;
    }
    
    // Add this token to the list with the timestamp it was created
    tokenList[tokenId] = {
        used: false,
        time: Date.now(),
        clientIp: null
    };

    logger.log('info', "Token (%s) added.", tokenId);
    return true;
}

function delToken(tokenId) {
    if (tokenId in tokenList) {
        delete tokenList[tokenId];
        logger.log('info', "Token (%s) removed.", tokenId);
        return true;
    }

    logger.log('warn', "Token (%s) not found!", tokenId);
    return false;
}

function listTokens() {
    var result = {
        size: 0,
        tokens: {}
    };

    // Get available tokens
    for (var token in tokenList) {
        result.tokens[result.size] = token;
        result.size++;
    }

    return result;
}

function addEmsToken(tokenId) {
    var index = findEmsToken(tokenId);
    if (index !== null) {
        logger.log('warn', "EMS token (%s) already exists!", tokenId);
        return false;
    }

    config.emsTokens.push(tokenId);

    logger.log('info', "EMS token (%s) added.", tokenId);
    return true;
}

function delEmsToken(tokenId) {
    if (config.emsTokens.length) {
        var index = findEmsToken(tokenId);
        if (index !== null) {
            // Remove this token from the list
            config.emsTokens.splice(index, 1);

            logger.log('info', "EMS token (%s) removed.", tokenId);
            return true;
        }
    }

    logger.log('warn', "EMS token (%s) not found!", tokenId);
    return false;
}

function listEmsTokens() {
    // List all EMS tokens available
    var result = {
        size: config.emsTokens.length,
        emsTokens: config.emsTokens
    };

    return result;
}

function findEmsToken(tokenId) {
    if (config.emsTokens.length) {
        for (var i = 0; i < config.emsTokens.length; i++) {
            if (config.emsTokens[i] === tokenId) {
                // Return the index of this room
                return i;
            }
        }
    }

    return null;
}

function addRoom(roomId) {
    var index = findRoom(roomId);
    if (index !== null) {
        logger.log('warn', "Room (%s) already exists!", roomId);
        return false;
    }

    config.allowedrooms.push(roomId);

    logger.log('info', "Room (%s) added.", roomId);
    return true;
}

function delRoom(roomId) {
    if (config.allowedrooms.length) {
        var index = findRoom(roomId);
        if (index !== null) {
            // Remove this room from the list
            config.allowedrooms.splice(index, 1);

            logger.log('info', "Room (%s) removed.", roomId);
            return true;
        }
    }

    logger.log('warn', "Room (%s) not found!", roomId);
    return false;
}

function listRooms() {
    var result = {
        size: 0,
        rooms: {}
    };

    // List all rooms available
    for (var room in roomTokens) {
        result.rooms[result.size] = {
            name: room,
            emsConnected: roomTokens[room].hasEms,
            emsIp: roomTokens[room].hasEms ? roomTokens[room].emsIp : null,
            tokens: roomTokens[room].tokens,
            allowed: (room in config.allowedrooms)
        };

        result.size++;
    }

    return result;
}

function version() {
    var result = {
        version: ersVersion
    };

    return result;
}

function findRoom(roomId) {
    if (config.allowedrooms.length) {
        for (var i = 0; i < config.allowedrooms.length; i++) {
            if (config.allowedrooms[i] === roomId) {
                // Return the index of this room
                return i;
            }
        }
    }

    return null;
}

function describeRoom(name) {
    var clients = io.sockets.clients(name);
    var result = {
        clients: {}
    };
    clients.forEach(function (client) {
        result.clients[client.id] = client.resources;
    });
    return result;
}

function clientsInRoom(name) {
    return io.sockets.clients(name).length;
}

function safeCb(cb) {
    if (typeof cb === 'function') {
        return cb;
    } else {
        return function () {
        };
    }
}

function leaveRoom(selectedClient) {
    if (selectedClient && selectedClient.room) {
        if (selectedClient.isEms) {
            // Clear the flag of the EMS
            roomTokens[selectedClient.room].hasEms = false;
        } else {
            // If this is a client, clear the token used on this room
            if (selectedClient.token in roomTokens[selectedClient.room].tokens) {
                delete roomTokens[selectedClient.room].tokens[selectedClient.token];
            }
        }

        selectedClient.leave(selectedClient.room);
        selectedClient.room = undefined;
    }
}

function initializeRoom(name, emsId) {
    if (emsId) {
        roomTokens[name] = {hasEms: true, emsId: emsId, tokens: {}};
    } else {
        roomTokens[name] = {hasEms: false, emsId: null, tokens: {}};
    }
}

io.sockets.on('connection', function (client) {
    client.resources = {
        screen: false,
        video: true,
        audio: false
    };

    logger.log('info', 'Connected.');

    // pass a message to another id
    client.on('message', function (details) {
        if (!details)
            return;

        var otherClient = io.sockets.sockets[details.to];
        if (!otherClient)
            return;

        details.from = client.id;
        otherClient.emit('message', details);

        logger.log('debug', 'Message received: %s', JSON.stringify(details));
    });

    client.on('join', join);

    function removeFeed() {
        // Notify connected client that we are leaving or disconnected
        var otherClient = io.sockets.sockets[client.partnerId];
        if (otherClient) {
            otherClient.emit('remove', {
                id: client.id,
                type: 'video'
            });
        }

        leaveRoom(client);
    }

    function joinRoom(name) {
        client.join(name);
        client.room = name;
    }

    function join(name, data, cb) {
        // Sanity check
        if (typeof name !== 'string')
            return;

        logger.log('debug', 'Data: %s', JSON.stringify(data));

        // Check if we allow this room name
        if (config.allowedrooms.length && (findRoom(name) === null)) {
            logger.log('error', "Room (%s) is invalid!", name);
            safeCb(cb)('invalidRoom');
            return;
        }

        var isEms = false;
        if ((data !== null) && (typeof data.EMS !== 'undefined')) {
            isEms = data.EMS;
        }

        // Check if we allow this EMS
        if (isEms) {
            if (config.emsTokens.length && (findEmsToken(data.emsToken) === null)) {
                logger.log('error', "EMS token (%s) is invalid!", data.emsToken);
                safeCb(cb)('invalidEmsToken');
                return;
            }
        }

        // Initialize the list to be returned
        var result = {
            clients: {}
        };

        var remoteIp = client.handshake.address.address;

        // Check if this is an EMS endpoint
        if (isEms) {
            // Check that the room name is not being used by another EMS
            var used = (name in roomTokens);
            if (used && roomTokens[name].hasEms) {
                logger.log('error', "Room (%s) is already taken.", name);
                safeCb(cb)('taken');
                return;
            } else if (used) {
                // This room has already been initialized, update the values
                roomTokens[name].hasEms = true;
                roomTokens[name].emsId = client.id;
            }
            else {
                // Initialize this room and indicate that it is taken
                initializeRoom(name, client.id);
            }

            // Add the IP address of this EMS
            roomTokens[name].emsIp = remoteIp;

            // Check if we have clients waiting, pick one and process
            var clients = io.sockets.clients(name);
            if (clients.length > 0) {
                // Pick the first
                var selectedClient = clients[0];

                // Populate the list
                result.clients[selectedClient.id] = selectedClient.resources;

                // Assign the partner IDs
                selectedClient.partnerId = client.id;
                client.partnerId = selectedClient.id;

                // Remove this client from the room
                leaveRoom(selectedClient);

                // Unset the flag that this room has an EMS since it did not join anyway
                roomTokens[name].hasEms = false;
            } else {
                // No client yet, join this room waiting...
                joinRoom(name);
            }

            // Indicate that this client is EMS
            client.isEms = true;
            client.token = null;
        } else {
            // In case that the client joined first without EMS yet, make sure that
            // details passed by client is not lost
            client.resources.data = data;

            // Validate the token
            var token = uuid();
            if (config.server.tokensEnabled) {
                if ((data === null) || (typeof data.token === 'undefined')) {
                    // Invalid token
                    logInvalidToken('empty', remoteIp);
                    safeCb(cb)('invalidToken');
                    return;
                }

                // Check if the passed token is valid
                if (!(data.token in tokenList)) {
                    // Token passed already expired or is invalid
                    logInvalidToken(data.token, remoteIp);
                    safeCb(cb)('invalidToken');
                    return;
                }
                
                // Check if this has been used already and that if it will be used
                // again, the same IP address is used
                if (tokenList[data.token].used) {
                    if (tokenList[data.token].clientIp !== remoteIp) {
                        logger.log('error', "Token (%s) requested by %s is already used by %s!", 
                        data.token, remoteIp, tokenList[data.token].clientIp);
                        safeCb(cb)('invalidToken');
                        return;
                    }
                }

                // Valid token, use it.
                token = data.token;

                // Mark as used
                tokenList[token].used = true;
                tokenList[token].clientIp = remoteIp;
            }

            // Check if this room exists
            if (!(name in roomTokens)) {
                initializeRoom(name, null);
            }

            // Leave any existing rooms
            leaveRoom(client);

            // Add the token
            client.token = token;

            // Assign this client to the room - tokens list
            if (typeof roomTokens[name].tokens[client.token] === 'undefined') {
                roomTokens[name].tokens[client.token] = {
                    id: client.id,
                    clientIp: remoteIp
                };
            } else {
                roomTokens[name].tokens[client.token].id = client.id;
                roomTokens[name].tokens[client.token].clientIp = remoteIp;
            }

            if (roomTokens[name].hasEms) {
                var emsId = roomTokens[name].emsId;

                // Populate the list to be returned
                result.clients[emsId] = client.resources;

                // Set the client Id on ems
                var ems = io.sockets.sockets[emsId];
                if (ems) {
                    ems.partnerId = client.id;
                }

                // Set the EMS Id on this client
                client.partnerId = emsId;

                // Remove EMS from this room
                leaveRoom(ems);

                // Pass any data to EMS
                var detail = {
                    id: client.id,
                    data: data
                };
                ems.emit('clientJoined', detail);
            } else {
                // Join this room, waiting for EMS...
                joinRoom(name);
            }
        }
        
        // Tag client with the room it 'joined'
        //client.room = name;

        safeCb(cb)(null, result);
        logger.log('info', 'Client %s joined %s.', client.id, name);
    }

    // we don't want to pass "leave" directly because the
    // event type string of "socket end" gets passed too.
    client.on('disconnect', function () {
        removeFeed();

        logger.log('info', 'Disconnected.');
    });

    client.on('leave', function () {
        removeFeed()();

        logger.log('info', 'Leave...');
    });

    // support for logging full webrtc traces to stdout
    // useful for large-scale error monitoring
    client.on('trace', function (data) {
        logger.log('debug', 'WebRTC trace: ', JSON.stringify(
                [data.type, data.session, data.prefix, data.peer, data.time, data.value]
                ));
    });

    // tell client about stun and turn servers and generate nonces
    client.emit('stunservers', config.stunservers || []);

    // create shared secret nonces for TURN authentication
    // the process is described in draft-uberti-behave-turn-rest
    var credentials = [];
    config.turnservers.forEach(function (server) {
        var hmac = crypto.createHmac('sha1', server.secret);
        // default to 86400 seconds timeout unless specified
        var username = Math.floor(new Date().getTime() / 1000) + (server.expiry || 86400) + "";
        hmac.update(username);
        credentials.push({
            username: username,
            credential: hmac.digest('base64'),
            url: server.url
        });
    });
    client.emit('turnservers', credentials);
});

if (config.uid)
    process.setuid(config.uid);

logger.log('info', "ERS running on port %s...", config.server.port);

if (config.server.webServerEnabled) {
    logger.log('info', 'Webserver port: ' + config.server.port);
}

if (config.server.adminAccessEnabled) {
    logger.log('info', 'Admin access port: ' + config.server.adminPort);
}
